/* drivers/video/epson13733fb.c
 *
 * (c) 2008 Siemens Enterprise Communications
 * (c) 2008 Karl Bittnar <karl.bittnar@siemens.com>
 * (c) 2008 Viktor Karell <viktor.karell@siemens.com>
 *
 * Driver for Epson S1D13733 framebuffer chip
 *
 * Adapted from
 *  linux/drivers/video/skeletonfb.c
 *  linux/drivers/video/epson1355fb.c
 *  linux/drivers/video/epson/s1d13xxxfb.c (2.4 driver by Epson)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 *
 * 2012-02-14  Modified by Emre Cantimur <emre.cantimur@siemens-enterprise.com>
 *             Changed IOMEM allocation of display buffer to get free space for
 *             H.263 encode functionality.
 *
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/fb.h>

#include <asm/io.h>

#include <video/epson13733.h>

#define PFX "epson13733fb: "

#if 0
#define dbg(fmt, args...) do { printk(KERN_INFO PFX fmt, ## args); } while(0)
#else
#define dbg(fmt, args...) do { } while (0)
#endif

//-----------------------------------------------------------------------------
//
// Local Definitions
//
//---------------------------------------------------------------------------
//    #define REG0022_INDIRECT_INTERFACE_MEMORY_ADDRESS_0     0x0022
//    #define REG0024_INDIRECT_INTERFACE_MEMORY_ADDRESS_1     0x0024
//    #define REG0028_INDIRECT_INTERFACE_MEMORY_DATA_PORT     0x0028
#define REG0034_LCD_INTERFACE_COMMAND                   0x0034
#define REG0036_LCD_INTERFACE_PARAMETER                 0x0036
#define REG0038_LCD_INTERFACE_STATUS                    0x0038
#define REG003A_LCD_INTERFACE_FRAME_TRANSFER            0x003A
#define MAIN_WINDOW_OFFSET                              0x0   // see main windows display start address registers
#define LOOP_LIMIT                                          1000     // prevent endless loops


//-----------------------------------------------------------------------------
//
// Function Prototypes
//
//-----------------------------------------------------------------------------

static u8 panelReadReg( u8 reg );
static void setupForWinTechPanel( struct epson13733fb_par *par,
			struct epson13733fb_regval *initregs,
			const unsigned int size );

extern int get_WP_type(void);
int panelReadType( void );
static int panelProgramWintech( void );
static void panelRefreshWintech( void );

//-----------------------------------------------------------------------------
//
// Local globals
//
//-----------------------------------------------------------------------------

S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);

static u8 panel_shadow_reg[64];
static struct fb_info *info=NULL;
static struct epson13733fb_regval *initregs_ptr=NULL;
static unsigned int initregs_size=0;
static int emv_work_run=0;
static struct semaphore emv_work_sem;
static struct delayed_work emv_work;


static inline u16
epson13733fb_readreg16(struct epson13733fb_par *par, u16 regno)
{
	u16 value;

	/* access s1d13733 registers only with 16-bit reads/writes */
        if (regno & 0x01)
	{
		dbg("_readreg16: corrected odd register 0x%04X\n", regno);
		regno &= ~0x0001;
	}
	value = ioread16(par->regs + regno);
	dbg("_readreg16(0x%04x): value=0x%04X\n", regno, value);
	return value;
}

static inline void
epson13733fb_writereg16(struct epson13733fb_par *par, u16 regno, u16 value)
{
	/* access s1d13733 registers only with 16-bit reads/writes */
        if (regno & 0x01)
	{
		dbg("_writereg16: corrected odd register 0x%04X\n", regno);
		regno &= ~0x0001;
	}
	dbg("_writereg16(0x%04x): value=0x%04X\n", regno, value);
	iowrite16(value, par->regs + regno);
}

static inline void
epson13733fb_writeDisplay16(char __iomem *screen_base, u32 addr, u32 value)
{
/*	dbg("_writeDisplay16(0x%04x): value=0x%04X\n", addr, value);*/
	*(u16 *)(screen_base + (addr & ~0x1)) = value;
}

int panel_reg_monitor( void )
{
	static int idx=0;
	u8 reg[2];
	int ret=0;

	reg[0] = panelReadReg(idx);
	reg[1] = panelReadReg(idx+0x20);

	if(reg[0] != panel_shadow_reg[idx]){
		printk(	"panel reg[%x] %x -> %x\n", idx, panel_shadow_reg[idx], reg[0]);
		ret++;
	}
	if(reg[1] != panel_shadow_reg[idx+0x20]){
		printk(	"panel reg[%x] %x -> %x\n", idx+0x20, panel_shadow_reg[idx+0x20], reg[1]);
		ret++;
	}
	idx++;
	if(idx > 0xd) idx =0;
	return ret;
}

static void emv_refresh_work(struct work_struct *ws)
{
	if (info!=NULL && info->par!=NULL && initregs_ptr!=NULL)
	{
		// Workaround for EMV
		if( (panelReadType() == 1) || (panelReadType() == 3) ) {
			if(panel_reg_monitor()){
				dbg("Refreshing panel...\n");
				panelRefreshWintech();
				if(panel_reg_monitor()){
					dbg("Register refresh not sufficient. Reset the Panel...\n");
					setupForWinTechPanel(info->par, initregs_ptr, initregs_size);
					panelRefreshWintech();
				}
			}
		}
	}
	down(&emv_work_sem);
	if(emv_work_run == 1)
		schedule_delayed_work(&emv_work, HZ);
	up(&emv_work_sem);
}

static inline void panelSPIDelay(u8 data)
{
//	dbg("panelSPIDelay %i\n", data);
	udelay(data);
}

static inline void panelSPIPanelWrite(u8 output)
{
//	dbg("panelSPIPanelWrite bc00001e 0x%02X\n", output);
	*FPGA_PAN_CNTRL = output;
	*FPGA_PAN_CNTRL = output;
}

static inline u8 panelSPIPanelRead(void)
{
	u8 output = *FPGA_PAN_CNTRL;
//	dbg("panelSPIPanelRead bc00001e 0x%02X\n", output);
	return output;
}

static inline u8 panelSPIPanelReadBit(void)
{
	u8 bit = (panelSPIPanelRead() & FPGA_PANEL_SPI_DATA_BIT) >> 3;

//	dbg("panelSPIPanelReadBit: %02X\n", bit);
	return bit;
}

static inline void panelSPI_CS_set(void)
{
	/* Idle SPI clock high */
	u8 output = FPGA_PANEL_SPI_EN_BIT | FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_CLK_BIT;
//	dbg("panelSPI_CS_set\n");
	panelSPIPanelWrite(output);
}

static inline void panelSPI_CS_clear(void)
{
	/* Idle SPI clock low */
	u8 output = FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_CLK_BIT;
//	dbg("panelSPI_CS_clear\n");
	panelSPIPanelWrite(output);
}

static void panelSPIBitWrite(u8 data)
{
	int i;
	u8 output;
	u8 base = FPGA_PANEL_DEFAULT_WRITE_BITS;

//	dbg("panelSPIBitWrite 0x%02X\n", data);
	for (i = 0; i < 8; i++)
	{
		/* MSB first */
		output = base | (((data << i) & 0x80) >> 4);
		panelSPIPanelWrite(output);
		output |= FPGA_PANEL_SPI_CLK_BIT;
		panelSPIPanelWrite(output);
	}
//	dbg("panelSPIBitWrite end\n");
}

static u8 panelSPIBitRead(void)
{
	int i;
	u8 output = 0;
	u8 base = FPGA_PANEL_DEFAULT_READ_BITS;

//	dbg("panelSPIBitRead start\n");
	for (i = 0; i < 8; i++)
	{
		/* MSB first */
		output <<= 1;
		panelSPIPanelWrite(base);
		panelSPIPanelWrite(base | FPGA_PANEL_SPI_CLK_BIT);
		output |= panelSPIPanelReadBit();
	}
//	dbg("panelSPIBitRead end: 0x%02X\n", output);
	return output;
}

static inline void panelSPIBitWriteReg(u8 reg)
{
//	dbg("panelSPIBitWriteReg 0x%02X\n", reg);
	/* When writing register set bit 2, the write bit */
	panelSPIBitWrite((reg << 2) | 0x2);
}

static inline void panelSPIBitReadReg(u8 reg)
{
//	dbg("panelSPIBitReadReg 0x%02X\n", reg);
	/* When writing register set bit 2, the write bit */
	panelSPIBitWrite(reg << 2);
}

static inline void panelSPISetDir(void)
{
	u8 output = FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_DIR_BIT | FPGA_PANEL_SPI_CLK_BIT;
//	dbg("panelSPISetDir\n");
	panelSPIPanelWrite(output);
}

static void panelWriteRealReg( u8 reg, u8 value)
{
//	dbg("panelWriteRealReg 0x%02X 0x%02X\n", reg, value);
	panel_shadow_reg[reg] = value;
	panelSPISetDir();
	panelSPIBitWriteReg( reg );
	panelSPIBitWrite( value );
	panelSPISetDir();
	panelSPIDelay(2);
}

static void panelWriteReg( u8 reg, u8 value)
{
//	dbg("panelWriteReg 0x%02X 0x%02X\n", reg, value);
	panel_shadow_reg[reg] = value;
	panelWriteRealReg( reg, value );
}

static u8 panelReadReg( u8 reg )
{
	u8 value;
//	dbg("panelReadReg 0x%02X start\n", reg);
	panelSPISetDir();
	panelSPIBitReadReg(reg);
	value = panelSPIBitRead();
	panelSPISetDir();
	panelSPIDelay(2);
//	dbg("panelReadReg 0x%02X end: 0x%02X\n", reg, value);
	return value;
}

u8 panelWinTekReadReg( u8 reg )
{

	return( panelReadReg( reg ) );
}

void panelWinTekWriteReg( u8 reg, u8 value)
{
	panelWriteReg(reg,value);
}

u16 epsonReadReg( u16 reg )
{
	return( epson13733fb_readreg16( info->par, reg ) );
//	return( readReg16( reg ) );
}

void epsonWriteReg( u16 reg, u16 value)
{
	epson13733fb_writereg16( info->par, reg, value );
//	writeReg16(reg,value);
}

static void panelRefreshWintech( void )
{
	int i;
	u8 value;
	for( i=0x0;i<0xd;i++)
	{
		value = panel_shadow_reg[i];
/*		printk("REG[%02x] = %02x\n",i, value ); */
		panelWriteReg(i, value);
	}

	for( i=0x20;i<0x2d;i++)
	{
		value = panel_shadow_reg[i];
/*		printk("REG[%02x] = %02x\n",i, value ); */
		panelWriteReg(i, value);
	}
}

static int panelProgramWintech( void )
{
	dbg("panelProgramWintech\n");

	dbg("configure panel master\n");
	panelWriteReg(0x00, 0x03);
	panelWriteReg(0x01, 0x82);
	panelWriteReg(0x02, 0x12);
	panelWriteReg(0x03, 0xcc);

	/*
	The new WinTek controler NT39403B has two significant changes from the previous version NT39403
	The first is that the value of DDLY programed by setting the Source Timing Delay Control Register must be greater than 0x27.
	This requires changes to Epson registers 0x40 LCD1 Horizontal Total Register and 0x44 LCD1 Horizontal Display Period Start Position Register to
	add the additional delay to the HSYNC control ine.

	The second is that the setting of the HDLY in the Gate Timing Delay Control Register gives a timing HDLY+2. In the pevious
	version this was HDLY+1
	*/

	if(panelReadType() == 3)
	{
		/* NT39403B */
		panelWriteReg(0x04, 0x2e);
		panelWriteReg(0x05, 0x0D);
	}
	else
	{
		/*
		NT39403
		These settings are those used in the previous version of the driver.
		The setting of the Gate Timing Delay Control Register [0x5] to 0x10 means that
		the top 2 lines of the display are missing. The correct value to display all lines would be 0x0E
		*/
		panelWriteReg(0x04, 0x27);
		panelWriteReg(0x05, 0x10);
	}

	panelWriteReg(0x06, 0x00);
	panelWriteReg(0x07, 0x30);
	panelWriteReg(0x08, 0x08);
	panelWriteReg(0x09, 0x40);
	panelWriteReg(0x0a, 0x88);
	panelWriteReg(0x0b, 0x88);
	panelWriteReg(0x0c, 0x20);
	panelWriteReg(0x0d, 0x20);

	dbg("configure panel slave\n");
	panelWriteReg(0x20, 0x03);
	panelWriteReg(0x21, 0x82);
	panelWriteReg(0x22, 0x12);
	panelWriteReg(0x23, 0xcc);

	if(panelReadType() == 3)
	{
		panelWriteReg(0x24, 0x2e);
		panelWriteReg(0x25, 0x0D);
	}
	else
	{
	panelWriteReg(0x24, 0x27);
		panelWriteReg(0x25, 0x10);
	}

	panelWriteReg(0x26, 0x00);
	panelWriteReg(0x27, 0x30);
	panelWriteReg(0x28, 0x08);
	panelWriteReg(0x29, 0x40);
	panelWriteReg(0x2a, 0x88);
	panelWriteReg(0x2b, 0x88);
	panelWriteReg(0x2c, 0x20);
	panelWriteReg(0x2d, 0x20);

//	mdelay(200);

	return 0;
}

static int panelProgramReadWintech( void )
{
	printk("echo panelProgramReadWintech\n");
	printk("echo Master\n");
	printk("reg %02x = %02x\n",0x00,panelReadReg(0x00));
	printk("reg %02x = %02x\n",0x01,panelReadReg(0x01));
	printk("reg %02x = %02x\n",0x02,panelReadReg(0x02));
	printk("reg %02x = %02x\n",0x03,panelReadReg(0x03));
	printk("reg %02x = %02x\n",0x04,panelReadReg(0x04));
	printk("reg %02x = %02x\n",0x05,panelReadReg(0x05));
	printk("reg %02x = %02x\n",0x06,panelReadReg(0x06));
	printk("reg %02x = %02x\n",0x07,panelReadReg(0x07));
	printk("reg %02x = %02x\n",0x08,panelReadReg(0x08));
	printk("reg %02x = %02x\n",0x09,panelReadReg(0x09));
	printk("reg %02x = %02x\n",0x0a,panelReadReg(0x0a));
	printk("reg %02x = %02x\n",0x0b,panelReadReg(0x0b));
	printk("reg %02x = %02x\n",0x0c,panelReadReg(0x0c));
	printk("reg %02x = %02x\n",0x0d,panelReadReg(0x0d));
	printk("reg %02x = %02x\n",0x0e,panelReadReg(0x0e));
	printk("echo SLAVE\n");
	printk("reg %02x = %02x\n",0x20,panelReadReg(0x20));
	printk("reg %02x = %02x\n",0x21,panelReadReg(0x21));
	printk("reg %02x = %02x\n",0x22,panelReadReg(0x22));
	printk("reg %02x = %02x\n",0x23,panelReadReg(0x23));
	printk("reg %02x = %02x\n",0x24,panelReadReg(0x24));
	printk("reg %02x = %02x\n",0x25,panelReadReg(0x25));
	printk("reg %02x = %02x\n",0x26,panelReadReg(0x26));
	printk("reg %02x = %02x\n",0x27,panelReadReg(0x27));
	printk("reg %02x = %02x\n",0x28,panelReadReg(0x28));
	printk("reg %02x = %02x\n",0x29,panelReadReg(0x29));
	printk("reg %02x = %02x\n",0x2a,panelReadReg(0x2a));
	printk("reg %02x = %02x\n",0x2b,panelReadReg(0x2b));
	printk("reg %02x = %02x\n",0x2c,panelReadReg(0x2c));
	printk("reg %02x = %02x\n",0x2d,panelReadReg(0x2d));
	printk("reg %02x = %02x\n",0x2e,panelReadReg(0x2e));
	return( 0 );
}

static int videoPanelType;

int panelReadType( void )
{
	return( videoPanelType );
}

static int panelType( int wp_type )
{
	u8	panel_cntrl_reg;
	u8	panel_system_control_reg;
	u8	panel_chip_version_reg;
	videoPanelType = 0;

	/* Check for hardware version */
	panelSPIPanelWrite( FPGA_PANEL_POWER_EN_BIT );
	if(wp_type == 3)
	{
		/*This is a WP3 display*/
		videoPanelType = 2;
	}
	else
	{
		panel_cntrl_reg = panelSPIPanelRead();
		printk("Panel Control %02x\n", panel_cntrl_reg);

		panelSPI_CS_set(); // This powers the panel and sets the control lines

		panel_system_control_reg = panelReadReg(0x00);
		panel_chip_version_reg     = panelReadReg(0x0e);

		if( panel_system_control_reg == 0xff )
		{
			videoPanelType = 0;
			/* panel_chip_version_reg should be 0xFF */
			panelSPIPanelWrite( 0 );
		}
		else if( panel_chip_version_reg == 0x00 )
		{
			/* panel_system_control_reg should be 0x3 */
			videoPanelType = 1;
		}
		else
		{
			/*
			   We have a non zero value in the version register.
			   It safer to default to videoPanelType 3 as these settings work for both displays
			   with the caveat that we lose the top line on the old wintech display

			  With the new wintek display the panel_system_control_reg would be 0x7 after reset
			  but will have been been set to 0x3 by uboot. The panel_chip_version_reg will be 0xC
			  for the current set of devices but this is subject to change by WinKec
			  */
			videoPanelType = 3;
		}
		printk("Panel Identification: system ctrl reg=%02x  chip version reg=%02x\n",panel_system_control_reg,panel_chip_version_reg);
	}
	return( videoPanelType );
}

static void epson_power_on_seq(struct epson13733fb_par *par)
{
	dbg("epson_power_on_seq\n");

	*FPGA_DISP_RESET = 0;	/* display controller reset */
	udelay(1000);		/* 1ms reset time - reset will be */
				/* in DCTRL_register_test() */

	/* Power on sequence - PLL set */

	/* PLL Power on sequence */
	epson13733fb_writereg16(par, 0x06,0x0003);	/* reset control */

	/* set clock configuration registers */
	epson13733fb_writereg16(par, 0x14, 0x0411);
	epson13733fb_writereg16(par, 0x12, 0x0001);
	epson13733fb_writereg16(par, 0x0C, 0x1027);
	epson13733fb_writereg16(par, 0x0E, 0x7390);
	epson13733fb_writereg16(par, 0x10, 0x0103);
	epson13733fb_writereg16(par, 0x18, 0x0000);
	epson13733fb_writereg16(par, 0x12, 0x0000);
	udelay(10000);					/* wait 10ms */

	epson13733fb_writereg16(par, 0x14, 0x0410);
}

static void setupForPhilipsPanel( struct epson13733fb_par *par,
			struct epson13733fb_regval *initregs,
			const unsigned int size )
{
    u32 i, timeout = LOOP_LIMIT;
    S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;
	printk("Configuring for Philips Panel\n");
    for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
    {
    s1dReg   = aS1DRegs[i].Index;
    s1dValue = aS1DRegs[i].Value;

    if (s1dReg == REG0034_LCD_INTERFACE_COMMAND || s1dReg == REG0036_LCD_INTERFACE_PARAMETER)
        while ((( epson13733fb_readreg16(par, REG0038_LCD_INTERFACE_STATUS) & 0x0001) != 0) && timeout--)
            continue;

    if (s1dReg == S1D_REGDELAYOFF || s1dReg == S1D_REGDELAYON || s1dReg == S1D_REGDELAYPLL)
        mdelay((int)s1dValue);
    else
        epson13733fb_writereg16( par, s1dReg, s1dValue);
    }
}


static void setupForWinTechPanel( struct epson13733fb_par *par,
			struct epson13733fb_regval *initregs,
			const unsigned int size )
{
	unsigned int i;
	if( videoPanelType == 1 )
        	printk("Configuring for Wintech NT3049 Controller with CCFL-backlight \n");
	else
        	printk("Configuring for Wintech NT3049B Controller with LED-backlight \n");

	epson_power_on_seq(par);
	panelSPISetDir();
	udelay(1000);
	for (i = 0; i < size; i++) {
		if( (initregs[i].reg == 0x44) && (panelReadType() == 3) )  // New Novatec Controller 2010
			initregs[i].value = 0x26;  /* start position 48 pixels */
		if( (initregs[i].reg == 0x40) && (panelReadType() == 3) )
			initregs[i].value = 0x55; /* total window 688 pixels   0x65 == 800 pixels*/
		epson13733fb_writereg16(par, initregs[i].reg, initregs[i].value);
	}
	udelay(1000);
	panelProgramWintech();

	mdelay(200);
//	panelProgramReadWintech();
}

static void setupForOptrixPanel( struct epson13733fb_par *par,
			struct epson13733fb_regval *initregs,
			const unsigned int size )
{
	unsigned int i;
	printk("Configuring for Optrex Panel\n");
	epson_power_on_seq(par);
	udelay(1000);

	for (i = 0; i < size; i++) {
		epson13733fb_writereg16(par, initregs[i].reg, initregs[i].value);
		}
	udelay(1000);
}

static inline void
epson13733fb_runinit(struct epson13733fb_par *par,
			struct epson13733fb_regval *initregs,
			const unsigned int size,
			int wp_type)
{
	dbg("_runinit called\n");

	*FPGA_DISP_RESET = 0;	/* display controller reset */

	switch( panelType(wp_type) )
	{
		case 0: setupForPhilipsPanel(par, initregs, size); break;
		case 1:
		case 3: setupForWinTechPanel(par, initregs, size);
				down(&emv_work_sem);
				if(emv_work_run == 0)
				{
					emv_work_run = 1;
					schedule_delayed_work(&emv_work, HZ / 40);
				}
				up(&emv_work_sem);
				break;
		case 2: setupForOptrixPanel(par, initregs, size); break;
	}
}

void epson13733fb_display_splash(struct epson13733fb_par *par,
		const struct epson13733fb_splash_data *splash_data, int wp_type,
		struct fb_info *info)
{
	char __iomem *screen_base = info->screen_base;
	u32 width = info->var.xres;
	u32 height = info->var.yres;
	int size = 1;
	u32 addr = 0;
	u32 startRow, endRow, startCol, endCol;
	u32 row, col, offset = 0;

	dbg("_display_splash: %i x %i for WP%i\n", width, height, wp_type);

	if (wp_type == 3) {
		size = 2;
	}
	else {
		size = 1;
	}
	startRow = (height - (splash_data->height/size)) / 2;
	endRow = startRow + (splash_data->height/size);

	startCol = (width - (splash_data->width/size)) / 2;
	endCol = startCol + (splash_data->width/size);


	dbg("startRow=%i endRow=%i startCol=%i endCol=%i size=%i\n",
		startRow, endRow, startCol, endCol, size);

	for (row = 0; row < height; row++)
	{
		if (row > startRow && row <= endRow)
		{
			offset += (splash_data->width * (size - 1));
		}
			for (col = 0; col < width; col++)
		{
			if (row < startRow || row >= endRow)
			{
				epson13733fb_writeDisplay16(screen_base, addr, 0xdf1c);
			}
			else
			{
				if (col >= startCol && col < endCol)
				{
					epson13733fb_writeDisplay16(screen_base, addr,
							splash_data->pixel_data[offset+=size]);
				}
				else
				{
					epson13733fb_writeDisplay16(screen_base, addr, 0xdf1c);
				}
			}
			addr += (S1D_DISPLAY_BPP/8);
		}
	}
}

void dummy_fb_fillrect(struct fb_info *p, const struct fb_fillrect *rect)
{
	dbg("_fb_fillrect called\n");
}

void dummy_fb_copyarea(struct fb_info *p, const struct fb_copyarea *area)
{
	dbg("_fb_copyarea called\n");
}

void dummy_fb_imageblit(struct fb_info *p, const struct fb_image *image)
{
	dbg("_fb_imageblit called\n");
}

/* framebuffer information structures */

static struct fb_ops epson13733fb_fbops = {
	.owner		= THIS_MODULE,

	/* to be replaced by any acceleration we can */
	.fb_fillrect	= dummy_fb_fillrect,
	.fb_copyarea	= dummy_fb_copyarea,
	.fb_imageblit	= dummy_fb_imageblit,
/*	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,*/
};

/*
 * Here we define the default struct fb_fix_screeninfo
 */
static struct fb_fix_screeninfo __devinitdata epson13733fb_fix = {
	.id		= S1D_FBID,
	.type		= FB_TYPE_PACKED_PIXELS,
	.visual		= FB_VISUAL_TRUECOLOR,
	.xpanstep	= 0,
	.ypanstep	= 0,
	.ywrapstep	= 0,
	.accel		= FB_ACCEL_NONE,
};

/*
 * Here we define the default struct fb_var_screeninfo
 */
static struct fb_var_screeninfo __devinitdata epson13733fb_var = {
	.xoffset	= 0,		/* offset from virtual to 	*/
	.yoffset	= 0,		/* visible resolution		*/

	.bits_per_pixel	= S1D_DISPLAY_BPP,
	.grayscale	= 0,		/* != 0 Graylevels instead of colors */

	.red		= {
		.offset		= 11,	/* bitfield in fb mem if true color, */
		.length		= 5,	/* else only length is significant */
		.msb_right	= 0,
	},

	.green		= {
		.offset		= 5,
		.length		= 6,
		.msb_right	= 0,
	},

	.blue		= {
		.offset		= 0,
		.length		= 5,
		.msb_right	= 0,
	},

	.transp		= {
		.offset		= 0,	/* transparency			*/
		.length		= 0,
		.msb_right	= 0,
	},

	.nonstd		= 0,		/* != 0 Non standard pixel format */
	.activate	= FB_ACTIVATE_NOW,
	.height		= -1,		/* height of picture in mm	*/
	.width		= -1,		/* width of picture in mm	*/
	.accel_flags	= 0,		/* (OBSOLETE) see fb_info.flags	*/

	/* Timing: All values in pixclocks, except pixclock (of course) */
	.pixclock	= S1D_DISPLAY_PCLK,	/* pixel clock in ps (pico seconds) */
	.left_margin	= 0,	/* time from sync to picture	*/
	.right_margin	= 0,	/* time from picture to sync	*/
	.upper_margin	= 0,	/* time from sync to picture	*/
	.lower_margin	= 0,
	.hsync_len	= 0,	/* length of horizontal sync	*/
	.vsync_len	= 0,	/* length of vertical sync	*/
	.sync		= 0,	/* see FB_SYNC_*		*/
	.vmode		= FB_VMODE_NONINTERLACED,
	.rotate		= 0,	/* angle we rotate counter clockwise */
};

/**
 *      epson13733fb_fill_screeninfo - Configure the fix and var
 *	screeninfo of the framebuffer of platform_device.
 *      @pdev: platform device
 *
 */
static void __devinit
epson13733fb_fill_screeninfo(struct platform_device *pdev)
{
	struct fb_info *info		= platform_get_drvdata(pdev);
	struct epson13733fb_pdata *pdata = pdev->dev.platform_data;
	struct fb_fix_screeninfo *fix	= &info->fix;
	struct fb_var_screeninfo *var	= &info->var;

	/* fill _fix_screeninfo */
	*fix = epson13733fb_fix;
	fix->mmio_start = pdev->resource[0].start;
	fix->mmio_len   = pdev->resource[0].end - pdev->resource[0].start + 1;
	fix->smem_start = pdev->resource[1].start + pdata->window_start_offset;
	fix->smem_len   = pdev->resource[1].end - (pdev->resource[1].start + pdata->window_start_offset) - 16 * 1024 + 1;

	/* fill _var_screeninfo */
	*var = epson13733fb_var;

	if (pdata->WP_type == 3)
	{
		/* This is a WP3 display */
		var->xres	= S1D_DISPLAY_WIDTH_WP3;	/* visible resolution */
		var->yres	= S1D_DISPLAY_HEIGHT_WP3;

		fix->line_length = S1D_DISPLAY_SCANLINE_BYTES_WP3;
	}
	else
	{
		/* This is a WP4 display */
		var->xres	= S1D_DISPLAY_WIDTH_WP4;
		var->yres	= S1D_DISPLAY_HEIGHT_WP4;

		fix->line_length = S1D_DISPLAY_SCANLINE_BYTES_WP4;
	}

	var->xres_virtual	= var->xres;	/* virtual resolution */
	var->yres_virtual	= var->yres;
}

static int __devexit
epson13733fb_remove(struct platform_device *pdev)
{
	struct fb_info *info = platform_get_drvdata(pdev);
	struct epson13733fb_par *par = NULL;

	dbg("_remove called: device is %p\n", pdev);

	if (info) {
		par = info->par;
		if (par && par->regs) {
			/* disable output & enable powersave */
			epson13733fb_writereg16(par, S1DREG_COM_DISP_MODE, 0x00);
			epson13733fb_writereg16(par, S1DREG_PS_CNF, 0x11);
			dbg("unmapping registers: %p\n", par->regs);
			iounmap(par->regs);
		}

		if (info->screen_base) {
			dbg("unmapping framebuffer: %p\n", info->screen_base);
			iounmap(info->screen_base);
		}

		dbg("releasing framebuffer: %p\n", info);
		framebuffer_release(info);
	}

	dbg("releasing resources\n");
	release_mem_region(pdev->resource[0].start,
			pdev->resource[0].end - pdev->resource[0].start + 1);
	release_mem_region(pdev->resource[1].start,
			pdev->resource[1].end - pdev->resource[1].start + 1);
	return 0;
}

void epson13733fb_init_video(void)
{
	dbg("_init_video called\n");
}

static int __devinit
epson13733fb_probe(struct platform_device *pdev)
{
	struct epson13733fb_par *default_par;
	struct epson13733fb_pdata *pdata = NULL;
	int ret = 0;
	u16 product_info;
	int wp_type = 0;
	unsigned int sram_for_display;

	dbg("_probe called: device is %p\n", pdev);

	printk(KERN_INFO "Epson S1D13733 FB Driver\n");

	/* evaluate WP_type: 3 (WP3) or 4 (WP4) */
	wp_type = get_WP_type();
	dbg("WP_type: %i\n", wp_type);
	if (wp_type != 3 && wp_type != 4) {
		dev_err(&pdev->dev, "invalid WP_type: %i (expected 3 or 4)\n",
			wp_type);
		ret = -ENODEV;
		goto bail;
	}

	/* platform-dependent hardware glue */
	dbg("platform_data: %p\n", pdev->dev.platform_data);
	if (!pdev->dev.platform_data) {
		dev_err(&pdev->dev, "platform data missing\n");
		ret = -EINVAL;
		goto bail;
	}
	pdata = pdev->dev.platform_data;
	pdata->WP_type = wp_type;

	dbg("num_resources: %i\n", pdev->num_resources);
	if (pdev->num_resources != 2) {
		dev_err(&pdev->dev, "invalid num_resources: %i\n",
			pdev->num_resources);
		ret = -ENODEV;
		goto bail;
	}

	/* resource[0] is registers, resource[1] is VRAM */
	if (pdev->resource[0].flags != IORESOURCE_MEM
			|| pdev->resource[1].flags != IORESOURCE_MEM) {
		dev_err(&pdev->dev, "invalid resource type\n");
		ret = -ENODEV;
		goto bail;
	}

	dbg("res0 (regs): %x (%u)\n",
	    pdev->resource[0].start,
	    pdev->resource[0].end - pdev->resource[0].start + 1);

	if (!request_mem_region(pdev->resource[0].start,
		pdev->resource[0].end - pdev->resource[0].start + 1, "epson13733 registers")) {
		dev_err(&pdev->dev, "request_mem_region failed\n");
		ret = -EBUSY;
		goto bail;
	}

	if (wp_type == 4) {
		sram_for_display = S1D_DISPLAY_WIDTH_WP4 * S1D_DISPLAY_HEIGHT_WP4 * S1D_DISPLAY_BPP / 8;
	} else {
		sram_for_display = S1D_DISPLAY_WIDTH_WP3 * S1D_DISPLAY_HEIGHT_WP3 * S1D_DISPLAY_BPP / 8;
	}

	/*
	 * The display are size should be multiple of PAGE_SIZE (mmap fails otherwise)
	 */
	sram_for_display += (sram_for_display % PAGE_SIZE);

	/* 16 KB for View resizer line buffer */
	pdata->window_start_offset = S1D_PHYSICAL_VMEM_SIZE - (sram_for_display + 16*1024);

	dbg("res1 (vmem): %x (%u)\n",
	    pdev->resource[1].start,
	    pdev->resource[1].end - pdev->resource[1].start + 1);

	if (!request_mem_region(pdev->resource[1].start,
		pdev->resource[1].end - pdev->resource[1].start + 1, "epson13733 framebuffer")) {
		dev_err(&pdev->dev, "request_mem_region failed\n");
		ret = -EBUSY;
		goto bail;
	}

	info = framebuffer_alloc(sizeof(struct epson13733fb_par), &pdev->dev);
	if (!info) {
		ret = -ENOMEM;
		goto bail;
	}

        /* add framebuffer to platform device */
	platform_set_drvdata(pdev, info);

	default_par = info->par;
	default_par->regs = ioremap_nocache(pdev->resource[0].start,
			pdev->resource[0].end - pdev->resource[0].start + 1);
	if (!default_par->regs) {
		dev_err(&pdev->dev, "unable to map registers\n");
		ret = -ENOMEM;
		goto bail;
	}

	info->pseudo_palette = default_par->pseudo_palette;

	info->screen_base = ioremap_nocache(pdev->resource[1].start,
			pdev->resource[1].end - pdev->resource[1].start - 1);

	if (!info->screen_base) {
		dev_err(&pdev->dev, "unable to map framebuffer\n");
		ret = -ENOMEM;
		goto bail;
	}
	info->screen_base += pdata->window_start_offset;

	product_info = epson13733fb_readreg16(default_par, S1DREG_PRODUCT_INFO);
        dbg("product_info: 0x%04X\n", product_info);
	if ((product_info >> 2) != S1D_PRODUCT_INFO) {
		dev_err(&pdev->dev, "epson13733 not found: 0x%04X\n", product_info);
//		ret = -ENODEV;
//		goto bail;
	}

	/* fill _fix_screeninfo and _var_screeninfo of framebuffer */
	epson13733fb_fill_screeninfo(pdev);

	printk(KERN_INFO PFX "regs mapped at 0x%p, fb %d KiB mapped at 0x%p\n",
		default_par->regs, info->fix.smem_len / 1024, info->screen_base);

	printk(KERN_INFO PFX "xres=%i yres=%i line_length=%i\n",
		info->var.xres, info->var.yres, info->fix.line_length);

	info->fbops = &epson13733fb_fbops;
	info->flags = FBINFO_DEFAULT;

	/* perform "manual" chip initialization, if needed */
	if (pdata && pdata->initregs_array && wp_type < pdata->initregs_array_size) {
		const struct epson13733fb_initreg *initregs = &pdata->initregs_array[wp_type];
		dbg("initregs_data=0x%p initregs_size: %i\n",
			initregs->initregs_data, initregs->initregs_size);
		if (initregs->initregs_data && initregs->initregs_size > 0)
			epson13733fb_runinit(info->par, initregs->initregs_data, initregs->initregs_size,
				wp_type);
	}

	/* display splash screen, if needed */
	if (pdata && pdata->splash_data)
		epson13733fb_display_splash(info->par, pdata->splash_data, wp_type, info);

	if (register_framebuffer(info) < 0) {
		dev_err(&pdev->dev, "register_framebuffer failed\n");
		ret = -EINVAL;
		goto bail;
	}

	epson13733fb_writereg16(default_par, 0x0210, pdata->window_start_offset & 0xffff);
	epson13733fb_writereg16(default_par, 0x0212, (pdata->window_start_offset & 0xf0000) >> 16);

	printk("************ pdata->window_start_offset: %x\n", pdata->window_start_offset);
	printk("************ REG[0x0210]: 0x%04x, REG[0x0212]: 0x%04x\n, REG[0x0218]: 0x%04x\n, REG[0x021A]: 0x%04x\n,", 
		 epson13733fb_readreg16(default_par,  0x0210), epson13733fb_readreg16(default_par,  0x0212),
		epson13733fb_readreg16(default_par,  0x0218), epson13733fb_readreg16(default_par,  0x021A));


	printk(KERN_INFO "fb%d: %s frame buffer device\n",
	       info->node, info->fix.id);

	return 0;

bail:
	epson13733fb_remove(pdev);
	return ret;
}


static struct platform_driver epson13733fb_driver = {
	.probe		= epson13733fb_probe,
	.remove		= epson13733fb_remove,
	.driver		= {
		.name	= S1D_DEVICENAME,
	},
};

static struct epson13733fb_regval aS1DRegs_WP3[] =
{
#include "LCD_WP3_regconf.h"
};

static struct epson13733fb_regval aS1DRegs_WP4[] =
{
#include "LCD_WP4_regconf.h"
};

#include "wpSplashData.h"

struct epson13733fb_initreg initregs_array[] = {
	[3] = {
		.initregs_data	= aS1DRegs_WP3,
		.initregs_size	= ARRAY_SIZE(aS1DRegs_WP3),
	},
	[4] = {
		.initregs_data	= aS1DRegs_WP4,
		.initregs_size	= ARRAY_SIZE(aS1DRegs_WP4),
	},
};

static struct epson13733fb_pdata epson13733fb_pdata = {
	.WP_type		= 0,
	.initregs_array		= initregs_array,
	.initregs_array_size	= ARRAY_SIZE(initregs_array),
	.splash_data		= (const struct epson13733fb_splash_data *)&wpSplashData,
	.platform_init_video	= epson13733fb_init_video,
};

static struct resource epson13733fb_resources[2] = {
	[0] = {	/* registers */
		.start	= S1D_PHYSICAL_REG_ADDR,
		.end	= S1D_PHYSICAL_REG_ADDR + S1D_PHYSICAL_REG_SIZE - 1,
		.name	= S1D_DEVICENAME "_registers",
		.flags	= IORESOURCE_MEM,
	},
	[1] = {	/* framebuffer */
		.start	= S1D_PHYSICAL_VMEM_ADDR, 
		.end	= S1D_PHYSICAL_VMEM_ADDR + S1D_PHYSICAL_VMEM_SIZE - 1,
		.name	= S1D_DEVICENAME "_framebuffer",
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device epson13733fb_device = {
	.name		= S1D_DEVICENAME,
	.id		= -1,
	.dev		= {
		.platform_data	= &epson13733fb_pdata,
	},
	.num_resources	= ARRAY_SIZE(epson13733fb_resources),
	.resource	= epson13733fb_resources,
};

static int __init
epson13733fb_init(void)
{
	int ret = 0;

	dbg("_init called\n");
	if (fb_get_options(S1D_DEVICENAME, NULL))
	{
        	dbg("fb_get_options failed\n");
		return -ENODEV;
	}

	INIT_DELAYED_WORK(&emv_work, emv_refresh_work);
	init_MUTEX(&emv_work_sem);

	ret = platform_driver_register(&epson13733fb_driver);
	dbg("platform_driver_register returns: %i\n", ret);

	if (!ret) {
		dbg("registering device %p\n", &epson13733fb_device);
		ret = platform_device_register(&epson13733fb_device);
		dbg("platform_device_register returns: %i\n", ret);
		if (ret)
			platform_driver_unregister(&epson13733fb_driver);
	}

	return ret;
}

module_init(epson13733fb_init);

#ifdef MODULE
static void __exit
epson13733fb_exit(void)
{
        dbg("_exit\n");
	/* stop the periodic work */
	down(&emv_work_sem);
	emv_work_run = 0;
	up(&emv_work_sem);
	platform_device_unregister(&epson13733fb_device);
	platform_driver_unregister(&epson13733fb_driver);
}

module_exit(epson13733fb_exit);
#endif

EXPORT_SYMBOL(panelWinTekWriteReg);
EXPORT_SYMBOL(panelWinTekReadReg);
EXPORT_SYMBOL(epsonReadReg);
EXPORT_SYMBOL(epsonWriteReg);
EXPORT_SYMBOL(panelReadType);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Framebuffer driver for Epson S1D13733 device");
MODULE_AUTHOR("Karl Bittnar <karl.bittnar@siemens.com>, Viktor Karell <viktor.karell@siemens.com>");
